""" Tests util functions """

import json

import pytest
from qibo.models.circuit import Circuit
from requests.models import Response

from qiboconnection.api_utils import deserialize_job_description
from qiboconnection.typings.enums import JobType
from qiboconnection.typings.responses.job_response import JobResponse
from qiboconnection.util import base64_decode, base64url_encode, from_kwargs, process_response


def test_base64url_encode():
    """Tests the base64url_encode() function."""
    payload = {
        "name": "alg001",
        "type": "Gate-Based Circuit",
        "options": {"number_qubits": 2, "initial_value": "zero"},
    }
    expected_encoded = "eyJuYW1lIjogImFsZzAwMSIsICJ0eXBlIjogIkdhdGUtQmFzZWQgQ2lyY3VpdCIsICJvcHRpb25zIjogeyJudW1iZXJfcXViaXRzIjogMiwgImluaXRpYWxfdmFsdWUiOiAiemVybyJ9fQ=="
    assert base64url_encode(payload) == expected_encoded


def test_base64url_encode_list():
    """Tests the base64url_encode() function providing it a list of dicts."""
    payload = [
        {
            "name": "alg001",
            "type": "Gate-Based Circuit",
            "options": {"number_qubits": 2, "initial_value": "zero"},
        }
    ]
    expected_encoded = "W3sibmFtZSI6ICJhbGcwMDEiLCAidHlwZSI6ICJHYXRlLUJhc2VkIENpcmN1aXQiLCAib3B0aW9ucyI6IHsibnVtYmVyX3F1Yml0cyI6IDIsICJpbml0aWFsX3ZhbHVlIjogInplcm8ifX1d"
    assert base64url_encode(json.dumps(payload)) == expected_encoded


def test_base64url_decode():
    """Test the base64url_decode() function"""
    data = "WzAuMSwgMC4xLCAwLjEsIDAuMSwgMC4xXQ=="
    expected_decoded = [0.1, 0.1, 0.1, 0.1, 0.1]
    assert json.loads(base64_decode(data)) == expected_decoded


def test_base64url_decode_list():
    """Test the base64url_decode() function providing it a list of dicts."""
    data = "W3sibmFtZSI6ICJhbGcwMDEiLCAidHlwZSI6ICJHYXRlLUJhc2VkIENpcmN1aXQiLCAib3B0aW9ucyI6IHsibnVtYmVyX3F1Yml0cyI6IDIsICJpbml0aWFsX3ZhbHVlIjogInplcm8ifX1d"
    expected_decoded = [
        {
            "name": "alg001",
            "type": "Gate-Based Circuit",
            "options": {"number_qubits": 2, "initial_value": "zero"},
        }
    ]
    assert json.loads(base64_decode(data)) == expected_decoded


def test_process_response(response: Response):
    """Test that process_response() recovers the correct parameters (text and status_code)."""

    processed_response = process_response(response=response)

    assert processed_response[0] == json.loads(response.text)
    assert processed_response[1] == response.status_code


def test_process_response_non_json(response_plain_text: Response):
    """Test that process_response() recovers the correct parameters (text and status_code)."""

    processed_response = process_response(response=response_plain_text)

    assert processed_response[0] == response_plain_text.text
    assert processed_response[1] == response_plain_text.status_code


def test_deserialize_job_description(
    base64_qibo_circuit: str, base64_qibo_circuits: str, base64_qililab_experiment: str
):
    """Unit test of deserialize_job_description()"""

    assert isinstance(
        deserialize_job_description(base64_description=base64_qibo_circuit, job_type=JobType.CIRCUIT), Circuit
    )
    assert isinstance(
        deserialize_job_description(base64_description=base64_qililab_experiment, job_type=JobType.EXPERIMENT), dict
    )

    assert deserialize_job_description(base64_description=base64_qibo_circuit, job_type="qiskit") is None
    assert isinstance(
        deserialize_job_description(base64_description=base64_qibo_circuits, job_type=JobType.CIRCUIT), list
    )


def test_from_kwargs():
    "Test of from_kwargs methods requires all explicitly typed attributes but accepts new ones."
    assert isinstance(
        from_kwargs(
            JobResponse,
            **{
                "user_id": 1,
                "device_id": 2,
                "description": "Job Description",
                "job_id": 1001,
                "queue_position": 5,
                "status": "Completed",
                "result": "Job Result",
                "number_shots": 10,
                "job_type": "whatever",
                "extra_arg": "Extra Argument",
            }
        ),
        JobResponse,
    )

    with pytest.raises(TypeError):
        isinstance(
            from_kwargs(
                JobResponse, **{"user_id": 1, "device_id": 2, "number_shots": 10, "extra_arg": "Extra Argument"}
            ),
            JobResponse,
        )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" SavedExperimentListingItem class """
# pylint: disable=too-many-instance-attributes
from dataclasses import dataclass, field
from datetime import datetime

from qiboconnection.typings.responses import SavedExperimentListingItemResponse
from qiboconnection.util import decode_jsonified_dict


@dataclass
class SavedExperimentListingItem:
    """SavedExperimentListing single item representation"""

    name: str
    user_id: int
    device_id: int
    description: str
    experiment: dict
    qililab_version: str
    id: int | None = field(default=None)
    created_at: datetime | None = field(default=None)

    @classmethod
    def from_response(cls, response: SavedExperimentListingItemResponse):
        """Constructor for SavedExperimentListingItems that takes in a SavedExperimentListingItemResponse"""
        return cls(
            id=response.id,
            created_at=response.created_at,
            name=response.name,
            user_id=response.user_id,
            device_id=response.device_id,
            description=response.description,
            experiment=decode_jsonified_dict(response.experiment),
            qililab_version=response.qililab_version,
        )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Runcard typing classes """
from dataclasses import dataclass


@dataclass
class RuncardRequest:
    """Base structure for a SavedExperiment body of any rest interaction"""

    name: str
    runcard: str
    device_id: int
    user_id: int
    description: str
    qililab_version: str



""" Test methods for Runcards classes """

from datetime import datetime, timezone

from qiboconnection.models.runcard import Runcard
from qiboconnection.typings.responses.runcard_response import RuncardRequest, RuncardResponse

# pylint: disable=no-member


def test_runcard_creation():
    """Tests Runcard Creation"""

    runcard = Runcard(
        name="runcard",
        description="description",
        user_id=0,
        device_id=0,
        runcard={"a": 0},
        qililab_version="0.0.0",
    )

    assert isinstance(runcard, Runcard)


def test_runcard_creation_from_response():
    """Tests Runcard Creation"""

    runcard_response = RuncardResponse(
        name="runcard",
        description="description",
        user_id=0,
        device_id=0,
        runcard="eyJhIjogMH0=",
        qililab_version="0.0.0",
        runcard_id=0,
        created_at=datetime.now(timezone.utc),
        updated_at=datetime.now(timezone.utc),
    )

    runcard = Runcard.from_response(response=runcard_response)

    assert isinstance(runcard, Runcard)
    assert runcard.runcard == {"a": 0}, "Decoded runcard does not coincide with expected"


def test_runcard_request():
    """Tests Runcard request builder method"""

    runcard = Runcard(
        name="runcard",
        description="description",
        user_id=0,
        device_id=0,
        runcard={"a": 0},
        qililab_version="0.0.0",
    )

    runcard_request = runcard.runcard_request()

    assert isinstance(runcard_request, RuncardRequest)
    assert runcard_request.runcard == "eyJhIjogMH0="



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" SavedExperiment class"""
from dataclasses import dataclass, field
from datetime import datetime

from qiboconnection.typings.requests import SavedExperimentRequest
from qiboconnection.typings.responses import SavedExperimentResponse
from qiboconnection.util import decode_jsonified_dict, jsonify_dict_and_base64_encode


@dataclass
class SavedExperiment:  # pylint: disable=too-many-instance-attributes
    """SavedExperiment representation"""

    name: str
    user_id: int
    device_id: int
    description: str
    experiment: dict
    results: dict
    qililab_version: str
    id: int | None = field(default=None)
    created_at: datetime | None = field(default=None)

    @property
    def _encoded_experiment(self):
        """return base64-encoded stringified jsonified experiment"""
        return jsonify_dict_and_base64_encode(self.experiment) if self.experiment is not None else None

    @property
    def _encoded_results(self):
        """return base64-encoded stringified jsonified results"""
        return jsonify_dict_and_base64_encode(self.results) if self.results is not None else None

    @classmethod
    def from_response(cls, response: SavedExperimentResponse):
        """SavedExperiment constructor that takes in an instance from a SavedExperimentResponse"""
        return cls(
            id=response.saved_experiment_id,
            created_at=response.created_at,
            name=response.name,
            description=response.description,
            user_id=response.user_id,
            device_id=response.device_id,
            experiment=decode_jsonified_dict(response.experiment),
            results=decode_jsonified_dict(response.results),
            qililab_version=response.qililab_version,
        )

    def saved_experiment_request(self, favourite: bool = False):
        """Created a SavedExperimentRequest instance"""
        return SavedExperimentRequest(
            name=self.name,
            user_id=self.user_id,
            device_id=self.device_id,
            description=self.description,
            experiment=self._encoded_experiment,
            results=self._encoded_results,
            favourite=favourite,
            qililab_version=self.qililab_version,
        )



""" Tests methods for Job """

from typing import cast

import numpy as np
import pytest
from qibo import gates
from qibo.models.circuit import Circuit

from qiboconnection.models import User
from qiboconnection.models.algorithm import (
    AlgorithmDefinition,
    AlgorithmName,
    AlgorithmOptions,
    AlgorithmType,
    InitialValue,
    ProgramDefinition,
)
from qiboconnection.models.devices import SimulatorDevice
from qiboconnection.models.devices.device import Device
from qiboconnection.models.job import Job
from qiboconnection.models.job_result import JobResult
from qiboconnection.typings.enums import JobStatus, JobType
from qiboconnection.typings.requests import JobRequest
from qiboconnection.typings.responses.job_response import JobResponse

from .data import simulator_device_inputs


@pytest.fixture(name="algorithm_definition")
def fixture_algorithm_definition() -> AlgorithmDefinition:
    """Create a new algorithm definition"""
    return AlgorithmDefinition(
        name=AlgorithmName.BELLSTATE,
        algorithm_type=AlgorithmType.GATE_BASED,
        options=AlgorithmOptions(number_qubits=2, initial_value=InitialValue.ZERO),
    )


@pytest.fixture(name="program_definition")
def fixture_program_definition(algorithm_definition: AlgorithmDefinition) -> ProgramDefinition:
    """Create a new program definition

    Returns:
        ProgramDefinition: Instance of the ProgramDefinition class.
    """
    return ProgramDefinition(algorithms=[algorithm_definition])


@pytest.fixture(name="circuit")
def fixture_circuit():
    """Create a circuit

    Returns:
        Circuit: Instance of Circuit with a couple of gates
    """
    circuit = Circuit(1)
    circuit.add(gates.H(0))
    circuit.add(gates.M(0))

    return circuit


@pytest.fixture(name="user")
def fixture_user() -> User:
    """Create a new user

    Returns:
        User: Instance of the User class.
    """
    return User(user_id=1, username="test-user", api_key="000-3333")


@pytest.fixture(name="simulator_device")
def fixture_simulator_device() -> SimulatorDevice:
    """Create a new SimulatorDevice

    Returns:
        SimulatorDevice: Instance of the SimulatorDevice class.
    """
    return SimulatorDevice(device_input=simulator_device_inputs[0])


def test_job_creation(circuit: Circuit, user: User, simulator_device: SimulatorDevice):
    """Test job creation

    Args:
        circuit (Circuit): Circuit
        user (User): User
        simulator_device (SimulatorDevice): SimulatorDevice
    """

    job_id = 23
    job_status = JobStatus.COMPLETED
    job_result = JobResult(
        job_id=job_id, http_response="WzAuMSwgMC4xLCAwLjEsIDAuMSwgMC4xXQ==", job_type=JobType.CIRCUIT
    )
    job = Job(
        circuit=circuit,
        user=user,
        device=cast(Device, simulator_device),
        job_status=job_status,
        job_result=job_result,
        id=job_id,
    )
    assert isinstance(job, Job)


def test_job_creation_default_values(circuit: Circuit, user: User, simulator_device: SimulatorDevice):
    """test job creation using the default values

    Args:
        circuit (Circuit): ProgramDefinition
        user (User): User
        simulator_device (SimulatorDevice): SimulatorDevice
    """

    job = Job(circuit=circuit, user=user, device=cast(Device, simulator_device))
    assert isinstance(job, Job)
    assert job.job_status == JobStatus.NOT_SENT
    assert job.job_result is None
    assert job.job_id == 0
    assert job.job_type == JobType.CIRCUIT
    assert job.user_id == user.user_id
    assert job.device_id == simulator_device.id
    with pytest.raises(ValueError) as e_info:
        job.result()
    assert e_info.value.args[0] == "Job result still not completed"
    with pytest.raises(ValueError) as e_info:
        _ = job.algorithms
    assert e_info.value.args[0] == "Job does not contains an algorithm Program"


def test_jobs_job_type_raises_value_error(circuit: Circuit, user: User, simulator_device: SimulatorDevice):
    """test job.job_type raises value error when unable to determine type

    Args:
        circuit (Circuit): circuit
        user (User): User
        simulator_device (SimulatorDevice): SimulatorDevice
    """

    job = Job(circuit=circuit, user=user, device=cast(Device, simulator_device))
    job.circuit = None

    with pytest.raises(ValueError) as e_info:
        _ = job.job_type

    assert e_info.value.args[0] == "Could not determine JobType"


def test_job_creation_experiment(user: User, simulator_device: SimulatorDevice):
    """test job creation using an experiment instead of a circuit

    Args:
        user (User): User
        simulator_device (SimulatorDevice): SimulatorDevice
    """

    job = Job(experiment={}, user=user, device=cast(Device, simulator_device))
    assert job.job_type == JobType.EXPERIMENT


def test_job_creation_experiment_raises_value_error_when_both_circuit_and_experiment_are_defined(
    circuit: Circuit, user: User, simulator_device: SimulatorDevice
):
    """test job creation using both experiment and circuit at the same time

    Args:
        circuit (Circuit): ProgramDefinition
        user (User): User
        simulator_device (SimulatorDevice): SimulatorDevice
    """

    with pytest.raises(ValueError) as e_info:
        _ = Job(experiment={}, circuit=circuit, user=user, device=cast(Device, simulator_device))
    assert e_info.value.args[0] == "Both circuit and experiment were provided, but execute() only takes one of them."


def test_job_creation_experiment_raises_value_error_when_neither_of_circuit_and_experiment_are_defined(
    user: User, simulator_device: SimulatorDevice
):
    """test job creation using neither experiment nor circuit

    Args:
        user (User): User
        simulator_device (SimulatorDevice): SimulatorDevice
    """
    with pytest.raises(ValueError) as e_info:
        _ = Job(experiment=None, circuit=None, user=user, device=cast(Device, simulator_device))
    assert e_info.value.args[0] == "Neither of experiment or circuit were provided,"


def test_job_request_with_circuit(circuit: Circuit, user: User, simulator_device: SimulatorDevice):
    """test job request

    Args:
        circuit (Circuit): Circuit
        user (User): User
        simulator_device (SimulatorDevice): SimulatorDevice
    """
    job_status = JobStatus.COMPLETED
    user_id = user.user_id
    job = Job(
        circuit=[circuit],
        user=user,
        device=cast(Device, simulator_device),
        job_status=job_status,
        id=23,
        nshots=10,
    )
    expected_job_request = JobRequest(
        user_id=user_id,
        device_id=simulator_device.id,
        description="['Ly8gR2VuZXJhdGVkIGJ5IFFJQk8gMC4xLjEyLmRldjAKT1BFTlFBU00gMi4wOwppbmNsdWRlICJxZWxpYjEuaW5jIjsKcXJlZyBxWzFdOwpjcmVnIHJlZ2lzdGVyMFsxXTsKaCBxWzBdOwptZWFzdXJlIHFbMF0gLT4gcmVnaXN0ZXIwWzBdOw==']",
        number_shots=10,
        job_type=JobType.CIRCUIT,
    )

    assert isinstance(job, Job)
    assert job.job_request == expected_job_request


def test_job_request_with_experiment(user: User, simulator_device: SimulatorDevice):
    """test job request

    Args:
        experiment(dict)
        user (User): User
        simulator_device (SimulatorDevice): SimulatorDevice
    """
    job_status = JobStatus.COMPLETED
    user_id = user.user_id
    job = Job(
        experiment={},
        user=user,
        device=cast(Device, simulator_device),
        job_status=job_status,
        id=23,
        nshots=10,
    )
    expected_job_request = JobRequest(
        user_id=user_id,
        device_id=simulator_device.id,
        description="e30=",
        number_shots=10,
        job_type=JobType.EXPERIMENT,
    )

    assert isinstance(job, Job)
    assert job.job_request == expected_job_request


def test_job_request_raises_value_error_if_not_circuit_or_experiment(
    circuit: Circuit, user: User, simulator_device: SimulatorDevice
):
    """test job raises proper exceptions when trying to build request with none of circuit, experiment

    Args:
        circuit (Circuit): Circuit
        user (User): User
        simulator_device (SimulatorDevice): SimulatorDevice
    """
    job = Job(
        circuit=circuit,
        user=user,
        device=cast(Device, simulator_device),
        job_status=JobStatus.COMPLETED,
        id=23,
        nshots=10,
    )
    job.circuit = None
    with pytest.raises(ValueError) as e_info:
        _ = job.job_request
    assert e_info.value.args[0] == "Could not determine JobType"


def test_job_request_raises_value_error_if_several_of_circuit_and_experiment(
    circuit: Circuit, user: User, simulator_device: SimulatorDevice
):
    """test job raises proper exceptions when trying to build request with more than one of circuit, experiment

    Args:
        circuit (Circuit): Circuit
        user (User): User
        simulator_device (SimulatorDevice): SimulatorDevice
    """
    job = Job(
        experiment={},
        user=user,
        device=cast(Device, simulator_device),
        job_status=JobStatus.COMPLETED,
        id=23,
        nshots=10,
    )
    job.circuit = circuit
    with pytest.raises(ValueError) as e_info:
        _ = job.job_request
    assert e_info.value.args[0] == "Could not determine JobType"


def test_update_with_job_response(circuit: Circuit, user: User, simulator_device: SimulatorDevice):
    """test update with job response

    Args:
        circuit (Circuit): Circuit
        user (User): User
        simulator_device (SimulatorDevice): SimulatorDevice
    """

    user_id = user.user_id
    job_status = JobStatus.PENDING
    job = Job(
        circuit=circuit,
        user=user,
        device=cast(Device, simulator_device),
        job_status=job_status,
        id=23,
    )
    job_response = JobResponse(
        user_id=user_id,
        number_shots=10,
        job_type=JobType.CIRCUIT,
        device_id=simulator_device.id,
        description="",
        job_id=job.id,
        queue_position=0,
        status=JobStatus.COMPLETED,
        result="gASVsAAAAAAAAACMFW51bXB5LmNvcmUubXVsdGlhcnJheZSMDF9yZWNvbnN0cnVjdJSTlIwFbnVtcHmUjAduZGFycmF5lJOUSwCFlEMBYpSHlFKUKEsBSwWFlGgDjAVkdHlwZZSTlIwCZjiUiYiHlFKUKEsDjAE8lE5OTkr_____Sv____9LAHSUYolDKAAAAAAAAPA_AAAAAAAA8D8AAAAAAADwPwAAAAAAAPA_AAAAAAAA8D-UdJRiLg==",
    )
    job.update_with_job_response(job_response=job_response)
    assert (job.result == np.array([1.0, 1.0, 1.0, 1.0, 1.0])).all()


def test_update_with_job_response_raises_error_when_updating_incorrect_job(
    circuit: Circuit, user: User, simulator_device: SimulatorDevice
):
    """test update with job response of different user or from different device raises ValueError

    Args:
        circuit (Circuit): Circuit
        user (User): User
        simulator_device (SimulatorDevice): SimulatorDevice
    """

    user_id = user.user_id
    job_status = JobStatus.PENDING
    job = Job(
        circuit=circuit,
        user=user,
        device=cast(Device, simulator_device),
        job_status=job_status,
        id=23,
    )

    job_response_different_user = JobResponse(
        user_id=user_id + 1 if user_id is not None else 1,
        number_shots=10,
        job_type=JobType.CIRCUIT,
        device_id=simulator_device.id,
        description="",
        job_id=job.id,
        queue_position=0,
        status=JobStatus.COMPLETED,
        result="WzAuMSwgMC4xLCAwLjEsIDAuMSwgMC4xXQ==",
    )
    with pytest.raises(ValueError) as e_info:
        job.update_with_job_response(job_response=job_response_different_user)
    assert e_info.value.args[0] == "Job response does not belong to the user."

    job_response_different_device = JobResponse(
        user_id=user_id,
        number_shots=10,
        job_type=JobType.CIRCUIT,
        device_id=simulator_device.id + 1,
        description="",
        job_id=job.id,
        queue_position=0,
        status=JobStatus.COMPLETED,
        result="WzAuMSwgMC4xLCAwLjEsIDAuMSwgMC4xXQ==",
    )
    with pytest.raises(ValueError) as e_info:
        job.update_with_job_response(job_response=job_response_different_device)
    assert e_info.value.args[0] == "Job response does not belong to the device."


# def test_job_creation_with_program(program_definition: ProgramDefinition, user: User, simulator_device: SimulatorDevice):
#     """Test job creation
#
#     Args:
#         program_definition (ProgramDefinition): ProgramDefinition
#         user (User): User
#         simulator_device (SimulatorDevice): SimulatorDevice
#     """
#
#     job_id = 23
#     job_status = JobStatus.COMPLETED
#     job_result = JobResult(job_id=job_id, http_response="WzAuMSwgMC4xLCAwLjEsIDAuMSwgMC4xXQ==")
#     job = Job(
#         program=program_definition,
#         user=user,
#         device=cast(Device, simulator_device),
#         job_status=job_status,
#         job_result=job_result,
#         id=job_id,
#     )
#     assert isinstance(job, Job)
#
#
# def test_job_creation_with_program_default_values(
#     program_definition: ProgramDefinition, user: User, simulator_device: SimulatorDevice
# ):
#     """test job creation using the default values
#
#     Args:
#         program_definition (ProgramDefinition): ProgramDefinition
#         user (User): User
#         simulator_device (SimulatorDevice): SimulatorDevice
#     """
#
#     job = Job(program=program_definition, user=user, device=cast(Device, simulator_device))
#     assert isinstance(job, Job)
#     assert job.job_status == JobStatus.NOT_SENT
#     assert job.job_result is None
#     assert job.job_id == 0
#
#
# def test_job_request_with_program(program_definition: ProgramDefinition, user: User, simulator_device: SimulatorDevice):
#     """test job request
#
#     Args:
#         program_definition (ProgramDefinition): ProgramDefinition
#         user (User): User
#         simulator_device (SimulatorDevice): SimulatorDevice
#     """
#     job_status = JobStatus.COMPLETED
#     job = Job(
#         program=program_definition,
#         user=user,
#         device=cast(Device, simulator_device),
#         job_status=job_status,
#         id=23,
#         nshots=10,
#     )
#     expected_job_request = JobRequest(
#         user_id=user.user_id,
#         device_id=simulator_device.id,
#         description="W3sibmFtZSI6ICJiZWxsLXN0YXRlIiwgInR5cGUiOiAiR2F0ZS1CYXNlZCBDaXJjdWl0IiwgIm9wdGlvbnMiOiB7Im51bWJlcl9xdWJpdHMiOiAyLCAiaW5pdGlhbF92YWx1ZSI6ICJ6ZXJvIn19XQ==",
#         number_shots=10,
#     )
#     assert isinstance(job, Job)
#     assert job.job_request == expected_job_request
#
#
# def test_update_with_job_response_and_program(program_definition: ProgramDefinition, user: User, simulator_device: SimulatorDevice):
#     """test update with job response
#
#     Args:
#         program_definition (ProgramDefinition): ProgramDefinition
#         user (User): User
#         simulator_device (SimulatorDevice): SimulatorDevice
#     """
#
#     job_status = JobStatus.PENDING
#     job = Job(
#         program=program_definition,
#         user=user,
#         device=cast(Device, simulator_device),
#         job_status=job_status,
#         id=23,
#     )
#     job_response = JobResponse(
#         user_id=user.user_id,
#         device_id=simulator_device.id,
#         description="",
#         job_id=job.id,
#         queue_position=0,
#         status=JobStatus.COMPLETED,
#         result="WzAuMSwgMC4xLCAwLjEsIDAuMSwgMC4xXQ==",
#     )
#     job.update_with_job_response(job_response=job_response)
#     assert job.result == [0.1, 0.1, 0.1, 0.1, 0.1]



""" Tests methods for Device """

import json

import pytest

from qiboconnection.models.devices import (
    Device,
    OfflineDevice,
    QuantumDevice,
    QuantumDeviceCharacteristics,
    SimulatorDevice,
    SimulatorDeviceCharacteristics,
)
from qiboconnection.models.devices.util import create_device, is_offline_device_input, is_quantum_device_input
from qiboconnection.typings.devices import (
    DeviceInput,
    OfflineDeviceInput,
    QuantumDeviceCharacteristicsInput,
    QuantumDeviceInput,
    SimulatorDeviceCharacteristicsInput,
    SimulatorDeviceInput,
)
from qiboconnection.typings.enums import DeviceStatus

from .data import (
    device_inputs,
    offline_device_inputs,
    quantum_device_characteristics_inputs,
    quantum_device_inputs,
    simulator_device_characteristics_inputs,
    simulator_device_inputs,
)

# pylint: disable=no-member


@pytest.mark.parametrize("device_input", device_inputs)
def test_device_constructor(device_input: DeviceInput):
    """Test device instance creation"""
    device = Device(device_input=device_input)
    assert isinstance(device, Device)
    assert device.name == device_input.device_name
    assert device.id == device_input.device_id


@pytest.mark.parametrize("device_input", device_inputs)
def test_device_string_representation(device_input: DeviceInput):
    """Tests Device().__str__() method"""
    device = Device(device_input=device_input)
    print(str(device))
    print(
        f"<Device: device_id={device._device_id},"
        f" device_name='{device._device_name}',"
        f" status='{device._status}',"
        f" availability='{device._availability}',"
        f" channel_id=None"
        f">"
    )

    assert (
        str(device) == f"<Device: device_id={device._device_id},"
        f" device_name='{device._device_name}',"
        f" status='{device._status}',"
        f" availability='{device._availability}',"
        f" channel_id={device._channel_id}"
        f">"
    )


@pytest.mark.parametrize("device_input", device_inputs)
def test_device_dict_representation(device_input: DeviceInput):
    """Tests Device().__dict__ property"""
    device = Device(device_input=device_input)
    expected_dict = {
        "device_id": device._device_id,
        "device_name": device._device_name,
        "status": device._status,
        "availability": device._availability,
    }
    assert device.__dict__ == expected_dict


@pytest.mark.parametrize("device_input", device_inputs)
def test_device_json_representation(device_input: DeviceInput):
    """Tests Device().toJSON() method"""
    device = Device(device_input=device_input)
    expected_dict = {
        "device_id": device._device_id,
        "device_name": device._device_name,
        "status": device._status,
        "availability": device._availability,
    }
    assert json.loads(device.toJSON()) == expected_dict


@pytest.mark.parametrize("simulator_device_input", simulator_device_inputs)
def test_simulator_device_constructor(simulator_device_input: SimulatorDeviceInput):
    """Tests SimulatorDevice class constructor"""
    assert isinstance(simulator_device_input, SimulatorDeviceInput)
    device = SimulatorDevice(device_input=simulator_device_input)
    assert isinstance(device, SimulatorDevice)


@pytest.mark.parametrize("simulator_device_input", simulator_device_inputs)
def test_simulator_device_dict_representation(simulator_device_input: SimulatorDeviceInput):
    """Tests SimulatorDevice().__str__() method"""
    device = SimulatorDevice(device_input=simulator_device_input)
    expected_dict = {
        "device_id": device._device_id,
        "device_name": device._device_name,
        "status": device._status,
        "availability": device._availability,
    }
    if device._characteristics:
        expected_dict |= {"characteristics": device._characteristics.__dict__}
    assert device.__dict__ == expected_dict


@pytest.mark.parametrize("simulator_device_input", simulator_device_inputs)
def test_simulator_device_json_representation(simulator_device_input: SimulatorDeviceInput):
    """Tests SimulatorDevice().toJSON() method"""
    device = SimulatorDevice(device_input=simulator_device_input)
    expected_dict = {
        "device_id": device._device_id,
        "device_name": device._device_name,
        "status": device._status,
        "availability": device._availability,
    }
    if device._characteristics:
        expected_dict |= {"characteristics": device._characteristics.__dict__}
    assert json.loads(device.toJSON()) == expected_dict


@pytest.mark.parametrize("simulator_device_characteristics_input", simulator_device_characteristics_inputs)
def test_simulator_device_characteristics_constructor(
    simulator_device_characteristics_input: SimulatorDeviceCharacteristicsInput,
):
    """Tests SimulatorCharacteristics class constructor"""
    characteristics = SimulatorDeviceCharacteristics(characteristics_input=simulator_device_characteristics_input)
    assert isinstance(characteristics, SimulatorDeviceCharacteristics)


@pytest.mark.parametrize("simulator_device_characteristics_input", simulator_device_characteristics_inputs)
def test_simulator_device_characteristics_json_representation(
    simulator_device_characteristics_input: SimulatorDeviceCharacteristicsInput,
):
    """Tests SimulatorDeviceCharacteristics().toJSON() method"""
    characteristics = SimulatorDeviceCharacteristics(characteristics_input=simulator_device_characteristics_input)
    expected_dict = {
        "type": characteristics._type,
        "cpu": characteristics._cpu,
        "gpu": characteristics._gpu,
        "os": characteristics._os,
        "kernel": characteristics._kernel,
        "ram": characteristics._ram,
    }
    assert characteristics.toJSON() == json.dumps(expected_dict, indent=2)


@pytest.mark.parametrize("quantum_device_input", quantum_device_inputs)
def test_quantum_device_constructor(quantum_device_input: QuantumDeviceInput):
    """Tests QuantumDevice class constructor"""
    assert isinstance(quantum_device_input, QuantumDeviceInput)
    device = QuantumDevice(device_input=quantum_device_input)
    assert isinstance(device, QuantumDevice)


@pytest.mark.parametrize("quantum_device_input", quantum_device_inputs)
def test_quantum_device_dict_representation(quantum_device_input: QuantumDeviceInput):
    """Tests QuantumDevice().__dict__ property"""
    device = QuantumDevice(device_input=quantum_device_input)
    expected_dict = {
        "device_id": device._device_id,
        "device_name": device._device_name,
        "status": device._status,
        "availability": device._availability,
    }
    if device._last_calibration_time is not None:
        expected_dict |= {"last_calibration_time": str(device._last_calibration_time)}
    if device._characteristics is not None:
        expected_dict |= {"characteristics": device._characteristics.__dict__}
    if device._calibration_details is not None:
        expected_dict |= {"calibration_details": device._calibration_details.__dict__}
    assert device.__dict__ == expected_dict


@pytest.mark.parametrize("quantum_device_input", quantum_device_inputs)
def test_quantum_device_json_representation(quantum_device_input: QuantumDeviceInput):
    """Tests QuantumDevice().toJSON() representation"""
    device = QuantumDevice(device_input=quantum_device_input)
    expected_dict = {
        "device_id": device._device_id,
        "device_name": device._device_name,
        "status": device._status,
        "availability": device._availability,
    }
    if device._last_calibration_time is not None:
        expected_dict |= {"last_calibration_time": str(device._last_calibration_time)}
    if device._characteristics is not None:
        expected_dict |= {"characteristics": device._characteristics.__dict__}
    if device._calibration_details is not None:
        expected_dict |= {"calibration_details": device._calibration_details.__dict__}
    assert json.loads(device.toJSON()) == expected_dict


@pytest.mark.parametrize("quantum_device_characteristics_input", quantum_device_characteristics_inputs)
def test_quantum_device_characteristics_constructor(
    quantum_device_characteristics_input: QuantumDeviceCharacteristicsInput,
):
    """Tests QuantumDeviceCharacteristics class constructor"""
    characteristics = QuantumDeviceCharacteristics(characteristics_input=quantum_device_characteristics_input)
    assert isinstance(characteristics, QuantumDeviceCharacteristics)


@pytest.mark.parametrize("quantum_device_characteristics_input", quantum_device_characteristics_inputs)
def test_quantum_device_characteristics_json_representation(
    quantum_device_characteristics_input: QuantumDeviceCharacteristicsInput,
):
    """Tests QuantumDeviceCharacteristics().toJSON() method"""
    characteristics = QuantumDeviceCharacteristics(characteristics_input=quantum_device_characteristics_input)
    expected_dict = {
        "type": characteristics._type,
    }
    assert json.loads(characteristics.toJSON()) == expected_dict


@pytest.mark.parametrize("offline_device_input", offline_device_inputs)
def test_offline_device_constructor(offline_device_input: OfflineDeviceInput):
    """Tests OfflineDevice class constructor"""
    device = OfflineDevice(device_input=offline_device_input)
    assert isinstance(device, OfflineDevice)


@pytest.mark.parametrize("offline_device_input", offline_device_inputs)
def test_is_offline_device_input(offline_device_input: OfflineDeviceInput):
    """Tests is_offline_device() utility function over an OfflineDevice-like input data"""
    assert is_offline_device_input(device_input=offline_device_input.__dict__)


@pytest.mark.parametrize("offline_device_input", offline_device_inputs)
def test_is_offline_device_input_rises_value_error_if_no_status(offline_device_input: OfflineDeviceInput):
    """Tests is_offline_device() utility function rises error when provided an OnlineDevice-like input data"""
    offline_device_input.status = None  # type: ignore

    with pytest.raises(ValueError) as e_info:
        _ = is_offline_device_input(device_input=offline_device_input.__dict__)

    assert e_info.value.args[0] == "'status' missing in device_input keys"


@pytest.mark.parametrize("quantum_device_input", quantum_device_inputs)
def test_is_quantum_device_input(quantum_device_input: QuantumDeviceInput):
    """Tests is_quantum_device_input() utility function over an QuantumDevice-like input data"""
    assert is_quantum_device_input(device_input=quantum_device_input.__dict__)


@pytest.mark.parametrize("offline_device_input", offline_device_inputs)
def test_create_offline_device(offline_device_input: OfflineDeviceInput):
    """Tests create_device() utility function creates an OfflineDevice when provided OfflineDevice-like input data"""
    offline_device_input.status = DeviceStatus.OFFLINE

    device = create_device(device_input=offline_device_input.__dict__)
    assert isinstance(device, OfflineDevice)


@pytest.mark.parametrize("simulator_device_input", simulator_device_inputs)
def test_create_simulator_device(simulator_device_input: SimulatorDeviceInput):
    """Tests create_device() utility function creates an SimulatorDevice when provided SimulatorDevice-like input
    data"""
    stripped_down_simulator_device_input = simulator_device_input.__dict__.copy()

    stripped_down_simulator_device_input["characteristics"] = stripped_down_simulator_device_input[
        "_characteristics"
    ].__dict__
    del stripped_down_simulator_device_input["_characteristics"]

    device = create_device(device_input=stripped_down_simulator_device_input)
    assert isinstance(device, SimulatorDevice)


@pytest.mark.parametrize("quantum_device_input", quantum_device_inputs)
def test_create_quantum_device(quantum_device_input: QuantumDeviceInput):
    """Tests create_device() utility function creates an QuantumDevice when provided QuantumDevice-like input data"""
    stripped_down_quantum_device_input = quantum_device_input.__dict__.copy()

    stripped_down_quantum_device_input["characteristics"] = stripped_down_quantum_device_input[
        "_characteristics"
    ].__dict__
    del stripped_down_quantum_device_input["_characteristics"]

    stripped_down_quantum_device_input["calibration_details"] = stripped_down_quantum_device_input[
        "_calibration_details"
    ].__dict__
    del stripped_down_quantum_device_input["_calibration_details"]

    device = create_device(device_input=stripped_down_quantum_device_input)
    assert isinstance(device, QuantumDevice)



""" Tests methods for Algorithms """

from qiboconnection.models.algorithm import (
    AlgorithmDefinition,
    AlgorithmName,
    AlgorithmOptions,
    AlgorithmType,
    InitialValue,
)


def test_algorithm_options_constructor():
    """Tests the AlgorithmOptions class constructor"""
    algorithm_options = AlgorithmOptions(number_qubits=2, initial_value=InitialValue.ZERO)
    assert isinstance(algorithm_options, AlgorithmOptions)


def test_algorithm_options_dict_representation():
    """Tests the AlgorithmOptions().__dict__ method"""
    algorithm_options = AlgorithmOptions(number_qubits=2, initial_value=InitialValue.ZERO)
    expected_dict = {"number_qubits": 2, "initial_value": InitialValue.ZERO.value}
    assert algorithm_options.__dict__ == expected_dict


def test_algorithm_definition_constructor():
    """Tests the AlgorithmDefinition class constructor"""
    algorithm_definition = AlgorithmDefinition(
        name=AlgorithmName.BELLSTATE,
        algorithm_type=AlgorithmType.GATE_BASED,
        options=AlgorithmOptions(number_qubits=2, initial_value=InitialValue.ZERO),
    )
    assert isinstance(algorithm_definition, AlgorithmDefinition)


def test_algorithm_definition_dict_representation():
    """Tests the AlgorithmDefinition().__dict__ method"""
    algorithm_definition = AlgorithmDefinition(
        name=AlgorithmName.BELLSTATE,
        algorithm_type=AlgorithmType.GATE_BASED,
        options=AlgorithmOptions(number_qubits=2, initial_value=InitialValue.ZERO),
    )
    expected_dict = {
        "name": AlgorithmName.BELLSTATE.value,
        "type": AlgorithmType.GATE_BASED.value,
        "options": AlgorithmOptions(number_qubits=2, initial_value=InitialValue.ZERO).__dict__,
    }
    assert algorithm_definition.__dict__ == expected_dict



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" PlottingResponse class """
from abc import ABC
from dataclasses import dataclass


@dataclass
class PlottingResponse(ABC):
    """Class for typecasting the PlottingService responses for requesting the creation of new plots."""

    websocket_url: str
    plot_id: int

    @classmethod
    def from_response(cls, websocket_url: str, plot_id: str):
        """Builds a PlottingResponse from a response json, where everything could be a string."""
        return cls(websocket_url=websocket_url, plot_id=int(plot_id))

    def to_dict(self):
        """Casts the info of the class as a dict."""
        return {"websocket_url": self.websocket_url, "plot_id": self.plot_id}
